import { Changesets } from './changesets';
import { CryptoUtils } from './crypto-utils';
import type { User } from './drizzle';
import { TimeOTP } from './users/totp';

export interface ChallengeBody<T> {
	aud: string;
	exp: number;
	data: T;
}

export class Challenge {
	static async issueChallenge<TChallenge>(
		challenge: TChallenge,
		recipient: string
	): Promise<string> {
		// 5-minute expiration on all challenges.
		const expiration = Date.now() + 5 * 60 * 1000;
		const body = <ChallengeBody<TChallenge>>{
			aud: recipient,
			exp: expiration,
			data: challenge
		};
		return CryptoUtils.encryptChallenge(body);
	}

	static async verifyChallenge<TRes>(
		challenge: string,
		recipient: string,
		code: string,
		secret: string
	) {
		const { aud, data, exp }: ChallengeBody<TRes> = await CryptoUtils.decryptChallenge(challenge);
		if (exp < Date.now() || aud !== recipient) {
			throw new Error('Invalid challenge');
		}

		if (!TimeOTP.validate(secret, code)) {
			throw new Error('Invalid token');
		}

		return data;
	}

	static async challengeFromBody<TRes>(
		body: FormData,
		recipient: string,
		secret: string
	): Promise<TRes | undefined> {
		const { challenge, otpCode } = Changesets.take<{ challenge?: string; otpCode?: string }>(
			['challenge', 'otpCode'],
			body
		);

		if (!challenge || !otpCode) {
			return;
		}

		return Challenge.verifyChallenge<TRes>(challenge, recipient, otpCode, secret);
	}

	static async authorizedChanges<TRes>(
		fields: (keyof TRes)[],
		body: FormData,
		subject: User
	): Promise<Partial<TRes>> {
		if (!body.has('challenge')) {
			return Changesets.take<TRes>(fields, body);
		}

		const userOtp = await TimeOTP.getUserOtp(subject);
		if (!userOtp) {
			throw new Error('User has no OTP');
		}

		const data = await Challenge.challengeFromBody<TRes>(body, subject.uuid, userOtp.token);
		if (!data) {
			throw new Error('Invalid request');
		}

		return data;
	}
}
