import { env } from '$env/dynamic/private';
import { readFile } from 'fs/promises';
import {
	SignJWT,
	exportJWK,
	importPKCS8,
	importSPKI,
	jwtVerify,
	type JWK,
	type KeyLike
} from 'jose';
import { join } from 'path';
import { v4 as uuidv4 } from 'uuid';

const { JWT_ALGORITHM, JWT_EXPIRATION, JWT_ISSUER } = env;

/**
 * Generate JWT keys using the following commands:
 * Private: openssl genpkey -out jwt.private.pem -algorithm RSA -pkeyopt rsa_keygen_bits:2048
 * Public:  openssl rsa -in jwt.private.pem -pubout -outform PEM -out jwt.public.pem
 */
export class JWT {
	static privateKey: KeyLike;
	static publicKey: KeyLike;
	static jwks: JWK;
	static jwksKid: string;

	static async init() {
		const privateKeyFile = await readFile(join('private', 'jwt.private.pem'), {
			encoding: 'utf-8'
		});
		const publicKeyFile = await readFile(join('private', 'jwt.public.pem'), { encoding: 'utf-8' });
		JWT.privateKey = await importPKCS8(privateKeyFile, JWT_ALGORITHM);
		JWT.publicKey = await importSPKI(publicKeyFile, JWT_ALGORITHM);
		JWT.jwks = await exportJWK(JWT.publicKey);
		JWT.jwksKid = uuidv4({ random: Buffer.from(JWT.jwks.n as string).subarray(0, 16) });
	}

	static async issue(claims: Record<string, unknown>, subject: string, audience?: string) {
		const sign = new SignJWT(claims)
			.setProtectedHeader({ alg: JWT_ALGORITHM })
			.setIssuedAt()
			.setSubject(subject)
			.setExpirationTime(JWT_EXPIRATION)
			.setIssuer(JWT_ISSUER);

		if (audience) {
			sign.setAudience(audience);
		}

		return sign.sign(JWT.privateKey);
	}

	static async verify(token: string, subject?: string, audience?: string) {
		const { payload } = await jwtVerify(token, JWT.publicKey, {
			issuer: JWT_ISSUER,
			subject,
			audience
		});
		return payload;
	}
}
