import type { UserSession } from '../../users';
import {
	InvalidRequest,
	UnsupportedResponseType,
	InvalidClient,
	UnauthorizedClient,
	InvalidScope,
	AccessDenied,
	InvalidGrant,
	InteractionRequired
} from '../error';
import {
	OAuth2AccessTokens,
	OAuth2Clients,
	OAuth2Codes,
	OAuth2Tokens,
	type CodeChallengeMethod
} from '../model';
import { OAuth2Users } from '../model/user';
import { OAuth2Response } from '../response';

export class OAuth2AuthorizationController {
	private static prehandle = async (url: URL, locals: App.Locals) => {
		if (!url.searchParams.has('redirect_uri')) {
			throw new InvalidRequest('redirect_uri field is mandatory for authorization endpoint');
		}

		const redirectUri = url.searchParams.get('redirect_uri') as string;
		// console.debug('Parameter redirect uri is', redirectUri);

		if (!url.searchParams.has('client_id')) {
			throw new InvalidRequest('client_id field is mandatory for authorization endpoint');
		}

		// Check for client_secret (prevent passing it)
		if (url.searchParams.has('client_secret')) {
			throw new InvalidRequest(
				'client_secret field should not be passed to the authorization endpoint'
			);
		}

		const clientId = url.searchParams.get('client_id') as string;
		// console.debug('Parameter client_id is', clientId);

		if (!url.searchParams.has('response_type')) {
			throw new InvalidRequest('response_type field is mandatory for authorization endpoint');
		}

		const responseType = url.searchParams.get('response_type') as string;
		// console.debug('Parameter response_type is', responseType);

		// Support multiple types
		const responseTypes = responseType.split(' ');
		let grantTypes: string[] = [];
		for (const responseType of responseTypes) {
			switch (responseType) {
				case 'code':
					grantTypes.push('authorization_code');
					break;
				case 'token':
					grantTypes.push('implicit');
					break;
				case 'id_token':
				case 'none':
					grantTypes.push(responseType);
					break;
				default:
					throw new UnsupportedResponseType('Unknown response_type parameter passed');
			}
		}

		// Filter out duplicates
		grantTypes = grantTypes.filter((value, index, self) => self.indexOf(value) === index);

		// "None" type cannot be combined with others
		if (grantTypes.length > 1 && grantTypes.includes('none')) {
			throw new InvalidRequest('Grant type "none" cannot be combined with other grant types');
		}

		// console.debug('Parameter grant_type is', grantTypes.join(' '));

		const client = await OAuth2Clients.fetchById(clientId);
		if (!client || client.activated === 0) {
			throw new InvalidClient('Client not found');
		}

		if (!(await OAuth2Clients.getRedirectUrls(client.client_id))?.length) {
			throw new UnsupportedResponseType('The client has not set a redirect uri');
		} else if (!(await OAuth2Clients.checkRedirectUri(client, redirectUri))) {
			throw new InvalidRequest('Wrong RedirectUri provided');
		}
		// console.debug('redirect_uri check passed');

		// The client needs to support all grant types
		for (const grantType of grantTypes) {
			if (grantType !== 'none' && !OAuth2Clients.checkGrantType(client, grantType)) {
				throw new UnauthorizedClient('This client does not support grant type ' + grantType);
			}
		}
		// console.debug('Grant type check passed');

		const scope = OAuth2Clients.transformScope(url.searchParams.get('scope') as string);
		if (!OAuth2Clients.checkScope(client, scope)) {
			throw new InvalidScope('Client does not allow access to this scope');
		}
		// console.debug('Scope check passed');

		const codeChallenge = url.searchParams.get('code_challenge') as string;
		const codeChallengeMethod =
			(url.searchParams.get('code_challenge_method') as CodeChallengeMethod) || 'plain';

		if (codeChallengeMethod && !OAuth2Tokens.challengeMethods.includes(codeChallengeMethod)) {
			throw new InvalidGrant('Invalid code challenge method');
		}

		return {
			client,
			user: locals.user,
			redirectUri,
			responseType,
			grantTypes,
			scope,
			codeChallenge,
			codeChallengeMethod
		};
	};

	private static posthandle = async (
		url: URL,
		{
			client,
			grantTypes,
			scope,
			user,
			codeChallenge,
			codeChallengeMethod,
			redirectUri,
			responseType
		}: Awaited<ReturnType<typeof OAuth2AuthorizationController.prehandle>>
	) => {
		let resObj: Record<string, string | number> = {};
		for (const grantType of grantTypes) {
			let data = null;
			switch (grantType) {
				case 'authorization_code':
					data = await OAuth2Codes.create(
						user.id,
						client.client_id,
						scope,
						OAuth2Tokens.codeTtl,
						url.searchParams.get('nonce') as string,
						codeChallenge,
						codeChallengeMethod
					);

					resObj = { code: data, ...resObj };

					break;
				case 'implicit':
					data = await OAuth2AccessTokens.create(
						user.id,
						client.client_id,
						scope,
						OAuth2Tokens.tokenTtl
					);

					resObj = {
						token_type: 'bearer',
						access_token: data,
						expires_in: OAuth2Tokens.tokenTtl,
						...resObj
					};

					break;
				case 'id_token':
					if (!scope.includes('openid')) {
						break;
					}

					data = await OAuth2Users.issueIdToken(
						user,
						client,
						scope,
						url.searchParams.get('nonce') as string | undefined
					);

					resObj = {
						id_token: data,
						...resObj
					};
					break;
				case 'none':
					resObj = {};
					break;
				default:
					throw new UnsupportedResponseType('Unknown response_type parameter passed');
			}
		}

		// Return non-code response types as fragment instead of query
		return OAuth2Response.responsePlain(url, resObj, redirectUri, responseType !== 'code');
	};

	static getRequest = async ({ locals, url }: { locals: App.Locals; url: URL }) => {
		const prehandle = await OAuth2AuthorizationController.prehandle(url, locals);
		const { client, scope, user } = prehandle;

		const prompt = ((url.searchParams.get('prompt') || '') as string).split(' ');
		let consented = false;

		// Check if the user has already consented to this client with this scope
		consented = await OAuth2Users.consented(user.id, client.client_id, scope);

		if (!consented && prompt.includes('none')) {
			throw new InteractionRequired('Interaction required!');
		}

		// Ask for consent
		if (
			!consented ||
			prompt.includes('login') ||
			prompt.includes('consent') ||
			prompt.includes('select_account')
		) {
			const sanitizedClient = await OAuth2Clients.authorizeClientInfo(client, scope);
			return {
				client: sanitizedClient,
				user: locals.session.data.user as UserSession
			};
		}

		return OAuth2AuthorizationController.posthandle(url, prehandle);
	};

	static actionRequest = async ({
		request,
		locals,
		url
	}: {
		locals: App.Locals;
		url: URL;
		request: Request;
	}) => {
		const prehandle = await OAuth2AuthorizationController.prehandle(url, locals);
		const { client, scope, user } = prehandle;

		const prompt = ((url.searchParams.get('prompt') || '') as string).split(' ');

		// Check if the user has already consented to this client with this scope
		const consented = await OAuth2Users.consented(user.id, client.client_id, scope);
		if (!consented && prompt.includes('none')) {
			throw new InteractionRequired('Interaction required!');
		}

		// Save consent
		if (!consented) {
			const body = await request.formData();
			if (!body?.has('decision')) {
				throw new InvalidRequest('No decision parameter passed');
			} else if (body.get('decision') === '0') {
				throw new AccessDenied('User denied access to the resource');
			}

			// console.debug('Decision check passed');

			await OAuth2Users.saveConsent(user, client, scope);
		}

		return OAuth2AuthorizationController.posthandle(url, prehandle);
	};
}
