import { ApiUtils } from '$lib/server/api-utils';
import { AccessDenied } from '../error';
import { OAuth2AccessTokens, OAuth2Tokens, type OAuth2AccessToken } from '../model';

export class OAuth2BearerController {
	static bearerFromRequest = async (
		request: Request,
		url: URL
	): Promise<OAuth2AccessToken | undefined> => {
		let token = null;

		// Look for token in header
		if (request.headers.has('authorization')) {
			const pieces = (request.headers.get('authorization') as string).split(' ', 2);

			// Check authorization header
			if (!pieces || pieces.length !== 2) {
				throw new AccessDenied('Wrong authorization header');
			}

			// Only bearer auth is supported
			if (pieces[0].toLowerCase() !== 'bearer') {
				throw new AccessDenied('Unsupported authorization method in header');
			}

			token = pieces[1];
		} else if (request.headers.has('x-access-token')) {
			token = request.headers.get('x-access-token');
		} else if (url?.searchParams.has('access_token')) {
			token = url.searchParams.get('access_token') as string;
		} else {
			const body = await ApiUtils.getJsonOrFormBody(request);
			if (!body.access_token) {
				throw new AccessDenied('Bearer token not found');
			}
			token = body.access_token;
		}

		// Try to fetch access token
		const object = await OAuth2AccessTokens.fetchByToken(token);
		if (!object || !OAuth2Tokens.checkTTL(object)) {
			throw new AccessDenied('Token not found or has expired');
		}
		return object;
	};
}
