import { ApiUtils } from '$lib/server/api-utils';
import { InvalidClient, InvalidRequest, UnauthorizedClient } from '../error';
import { OAuth2AccessTokens, OAuth2Clients } from '../model';
import { OAuth2Response } from '../response';

export class OAuth2IntrospectionController {
	static postHandler = async ({ request, url }: { request: Request; url: URL }) => {
		const body = await ApiUtils.getJsonOrFormBody(request);

		let clientId: string | null = null;
		let clientSecret: string | null = null;

		if (body.client_id && body.client_secret) {
			clientId = body.client_id as string;
			clientSecret = body.client_secret as string;
			// console.debug('Client credentials parsed from body parameters ', clientId, clientSecret);
		} else {
			if (!request.headers?.has('authorization')) {
				throw new InvalidRequest('No authorization header passed');
			}

			let pieces = (request.headers.get('authorization') as string).split(' ', 2);
			if (!pieces || pieces.length !== 2) {
				throw new InvalidRequest('Authorization header is corrupted');
			}

			if (pieces[0] !== 'Basic') {
				throw new InvalidRequest(`Unsupported authorization method: ${pieces[0]}`);
			}

			pieces = Buffer.from(pieces[1], 'base64').toString('ascii').split(':', 2);
			if (!pieces || pieces.length !== 2) {
				throw new InvalidRequest('Authorization header has corrupted data');
			}

			clientId = pieces[0];
			clientSecret = pieces[1];
			// console.debug('Client credentials parsed from basic auth header: ', clientId, clientSecret);
		}

		if (!body.token) {
			throw new InvalidRequest('Token not provided in request body');
		}

		const client = await OAuth2Clients.fetchById(clientId);

		if (!client || client.activated === 0) {
			throw new InvalidClient('Client not found');
		}

		const valid = OAuth2Clients.checkSecret(client, clientSecret);
		if (!valid) {
			throw new UnauthorizedClient('The client authentication was invalid');
		}

		const token = await OAuth2AccessTokens.fetchByToken(body.token);
		if (!token || token.clientIdPub !== clientId || token.expires_at.getTime() < Date.now()) {
			return OAuth2Response.response(url, { active: false });
		}

		const resObj = {
			active: true,
			scope: token.scope || '',
			client_id: clientId,
			exp: Math.floor(token.expires_at.getTime() / 1000)
		};

		return OAuth2Response.response(url, resObj);
	};
}
