import { ApiUtils } from '$lib/server/api-utils';
import {
	InvalidRequest,
	InvalidClient,
	UnauthorizedClient,
	OAuth2Error,
	UnsupportedGrantType,
	ServerError
} from '../error';
import { OAuth2Clients } from '../model';
import { OAuth2Response, type OAuth2TokenResponse } from '../response';
import * as tokens from './tokens';

export class OAuth2TokenController {
	static postHandler = async ({ url, request }: { url: URL; request: Request }) => {
		let clientId: string | null = null;
		let clientSecret: string | null = null;
		let grantType: string | null = null;

		const body = await ApiUtils.getJsonOrFormBody(request);

		if (body.client_id) {
			clientId = body.client_id as string;
			clientSecret = body.client_secret as string;
			// console.debug('Client credentials parsed from body parameters', clientId, clientSecret);
		} else {
			if (!request.headers?.has('authorization')) {
				throw new InvalidRequest('No authorization header passed');
			}

			let pieces = (request.headers.get('authorization') as string).split(' ', 2);
			if (!pieces || pieces.length !== 2) {
				throw new InvalidRequest('Authorization header is corrupted');
			}

			if (pieces[0] !== 'Basic') {
				throw new InvalidRequest(`Unsupported authorization method: ${pieces[0]}`);
			}

			pieces = Buffer.from(pieces[1], 'base64').toString('ascii').split(':', 2);
			if (!pieces || pieces.length !== 2) {
				throw new InvalidRequest('Authorization header has corrupted data');
			}

			clientId = pieces[0];
			clientSecret = pieces[1];
			// console.debug('Client credentials parsed from basic auth header:', clientId, clientSecret);
		}

		if (!body.grant_type) {
			throw new InvalidRequest('Request body does not contain grant_type parameter');
		}

		grantType = body.grant_type as string;
		// console.debug('Parameter grant_type is', grantType);

		// The spec does not allow using this grant type directly by this name,
		// but for verification purposes, we will simplify it below.
		if (grantType === 'device_code') {
			throw new UnsupportedGrantType('Grant type does not match any supported type');
		}

		if (grantType === 'urn:ietf:params:oauth:grant-type:device_code') {
			grantType = 'device_code';
		}

		const client = await OAuth2Clients.fetchById(clientId);
		if (!client || client.activated === 0) {
			throw new InvalidClient('Client not found');
		}

		// Device code flow does not require client authentication
		if (grantType !== 'device_code' || clientSecret) {
			const valid = OAuth2Clients.checkSecret(client, clientSecret);
			if (!valid) {
				throw new UnauthorizedClient('Invalid client secret');
			}
		}

		if (!OAuth2Clients.checkGrantType(client, grantType) && grantType !== 'refresh_token') {
			throw new UnauthorizedClient('Invalid grant type for the client');
		}
		// console.debug('Grant type check passed');

		let tokenResponse: OAuth2TokenResponse = {};
		try {
			switch (grantType) {
				case 'device_code':
					tokenResponse = await tokens.device(client, body.device_code);
					break;
				case 'authorization_code':
					tokenResponse = await tokens.authorizationCode(client, body.code, body.code_verifier);
					break;
				case 'client_credentials':
					tokenResponse = await tokens.clientCredentials(client, body.scope);
					break;
				case 'refresh_token':
					tokenResponse = await tokens.refreshToken(client, body.refresh_token);
					break;
				default:
					throw new UnsupportedGrantType('Grant type does not match any supported type');
			}

			if (tokenResponse) {
				return OAuth2Response.response(url, tokenResponse);
			}
		} catch (e) {
			return OAuth2Response.error(url, e as OAuth2Error);
		}

		throw new ServerError('Internal error');
	};
}
