import { CryptoUtils } from '$lib/server/crypto-utils';
import type { OAuth2Client, User } from '$lib/server/drizzle';
import { Users } from '$lib/server/users';
import { InvalidRequest, ServerError, InvalidGrant } from '../../error';
import {
	OAuth2AccessTokens,
	OAuth2Clients,
	OAuth2Codes,
	OAuth2RefreshTokens,
	OAuth2Tokens,
	type OAuth2Code
} from '../../model';
import { OAuth2Users } from '../../model/user';
import type { OAuth2TokenResponse } from '../../response';

/**
 * Issue an access token by authorization code
 * @param client - OAuth2 client
 * @param providedCode - Authorization code
 * @param codeVerifier - PCKE check verifier
 * @returns Access token.
 */
export async function authorizationCode(
	client: OAuth2Client,
	providedCode: string,
	codeVerifier?: string
): Promise<OAuth2TokenResponse> {
	const respObj: OAuth2TokenResponse = {
		token_type: 'bearer'
	};

	let code: OAuth2Code | undefined = undefined;

	if (!providedCode) {
		throw new InvalidRequest('code is mandatory for authorization_code grant type');
	}

	try {
		code = await OAuth2Codes.fetchByCode(providedCode);
	} catch (err) {
		console.error(err);
		throw new ServerError('Failed to call code.fetchByCode function');
	}

	if (code) {
		if (code.clientId !== client.id) {
			throw new InvalidGrant('Code was issued by another client');
		}

		if (!OAuth2Tokens.checkTTL(code)) {
			throw new InvalidGrant('Code has already expired');
		}
	} else {
		throw new InvalidGrant('Code not found');
	}

	// console.debug('Code fetched', code);

	const scope = code.scope || '';
	const cleanScope = OAuth2Clients.transformScope(scope);
	const userId = code.userId as number;
	const clientId = client.client_id;

	if (OAuth2Codes.getCodeChallenge) {
		const { challenge, method } = OAuth2Codes.getCodeChallenge(code);

		if (challenge && method) {
			if (!codeVerifier) {
				throw new InvalidGrant('Code verifier is required!');
			}

			if (method === 'plain' && !CryptoUtils.safeCompare(challenge, codeVerifier)) {
				throw new InvalidGrant('Invalid code verifier!');
			}

			if (
				method === 'S256' &&
				!CryptoUtils.safeCompare(
					CryptoUtils.createS256(codeVerifier),
					CryptoUtils.sanitizeS256(challenge)
				)
			) {
				throw new InvalidGrant('Invalid code verifier!');
			}
		}

		// console.debug('Code passed PCKE check');
	}

	if (!OAuth2Clients.checkGrantType(client, 'refresh_token')) {
		// console.debug('Client does not allow grant type refresh_token, skip creation');
	} else {
		try {
			respObj.refresh_token = await OAuth2RefreshTokens.create(userId, clientId, scope);
		} catch (err) {
			console.error(err);
			throw new ServerError('Failed to call refreshTokens.create function');
		}
	}

	try {
		respObj.access_token = await OAuth2AccessTokens.create(
			userId,
			clientId,
			scope,
			OAuth2Tokens.tokenTtl
		);
	} catch (err) {
		console.error(err);
		throw new ServerError('Failed to call accessTokens.create function');
	}

	respObj.expires_in = OAuth2Tokens.tokenTtl;
	// console.debug('Access token saved:', respObj.access_token);

	try {
		await OAuth2Codes.removeByCode(providedCode);
	} catch (err) {
		console.error(err);
		throw new ServerError('Failed to call codes.removeByCode function');
	}

	if (cleanScope.includes('openid')) {
		const user = await Users.getById(userId);

		try {
			respObj.id_token = await OAuth2Users.issueIdToken(
				user as User,
				client,
				cleanScope,
				code.nonce || undefined
			);
		} catch (err) {
			console.error(err);
			throw new ServerError('Failed to issue an ID token');
		}
	}

	return respObj;
}
