import type { OAuth2Client, User } from '$lib/server/drizzle';
import { Users } from '$lib/server/users';
import { InvalidRequest, ServerError, InvalidGrant, InvalidClient } from '../../error';
import {
	type OAuth2RefreshToken,
	type OAuth2AccessToken,
	OAuth2RefreshTokens,
	OAuth2AccessTokens,
	OAuth2Tokens
} from '../../model';
import type { OAuth2TokenResponse } from '../../response';

/**
 * Get a new access token from a refresh token. Scope change may not be requested.
 * @param client - OAuth2 client
 * @param providedToken - Refresh token
 * @returns Access token
 */
export async function refreshToken(
	client: OAuth2Client,
	providedToken: string
): Promise<OAuth2TokenResponse> {
	let user: User | undefined = undefined;
	let refreshToken: OAuth2RefreshToken | undefined = undefined;
	let accessToken: OAuth2AccessToken | undefined = undefined;

	const resObj: OAuth2TokenResponse = {
		token_type: 'bearer'
	};

	if (!providedToken) {
		throw new InvalidRequest('refresh_token is mandatory for refresh_token grant type');
	}

	try {
		refreshToken = await OAuth2RefreshTokens.fetchByToken(providedToken);
	} catch (err) {
		throw new ServerError('Failed to call refreshToken.fetchByToken function');
	}

	if (!refreshToken || !OAuth2Tokens.checkTTL(refreshToken)) {
		throw new InvalidGrant('Refresh token not found or it is already expired');
	}

	if (refreshToken.clientId !== client.id) {
		console.warn(
			'Client %s tried to fetch a refresh token which belongs to client %s!',
			client.id,
			refreshToken.clientId
		);
		throw new InvalidGrant('Refresh token not found');
	}

	try {
		user = await Users.getById(refreshToken.userId as number);
	} catch (err) {
		throw new InvalidClient('User not found');
	}

	if (!user) {
		throw new InvalidClient('User not found');
	}

	try {
		accessToken = await OAuth2AccessTokens.fetchByUserIdClientId(user.id, client.client_id);
	} catch (err) {
		throw new ServerError('Failed to call accessToken.fetchByUserIdClientId function');
	}

	// Remove old token
	if (accessToken && OAuth2Tokens.checkTTL(accessToken)) {
		await OAuth2Tokens.remove(accessToken);
	}

	// Remove old refresh token
	await OAuth2Tokens.remove(refreshToken);

	try {
		resObj.refresh_token = await OAuth2RefreshTokens.create(
			user.id,
			client.client_id,
			refreshToken.scope || ''
		);
	} catch (err) {
		throw new ServerError('Failed to call refreshTokens.create function');
	}

	try {
		resObj.access_token = await OAuth2AccessTokens.create(
			user.id,
			client.client_id,
			refreshToken.scope || '',
			OAuth2Tokens.tokenTtl
		);
	} catch (err) {
		throw new ServerError('Failed to call accessTokens.create function');
	}

	resObj.expires_in = OAuth2Tokens.tokenTtl;

	return resObj;
}
