import {
	DB,
	oauth2Client,
	oauth2ClientAuthorization,
	oauth2ClientUrl,
	type OAuth2Client,
	type User
} from '$lib/server/drizzle';
import { Users } from '$lib/server/users';
import { and, eq } from 'drizzle-orm';
import { OAuth2Clients } from './client';
import { OAuth2Tokens } from './tokens';
import { env } from '$env/dynamic/public';
import { JWT } from '$lib/server/jwt';

export class OAuth2Users {
	static async fetchFromLocals(locals: App.Locals) {
		return Users.getBySession(locals.session.data?.user);
	}

	static async consented(userId: number, clientId: string, scopes: string | string[]) {
		const normalized = OAuth2Clients.splitScope(scopes);
		return !!(
			await DB.drizzle
				.select({
					id: oauth2ClientAuthorization.id,
					scope: oauth2ClientAuthorization.scope
				})
				.from(oauth2ClientAuthorization)
				.innerJoin(oauth2Client, eq(oauth2ClientAuthorization.clientId, oauth2Client.id))
				.where(
					and(
						eq(oauth2Client.client_id, clientId),
						eq(oauth2ClientAuthorization.userId, userId),
						eq(oauth2ClientAuthorization.current, 1)
					)
				)
		).filter(({ scope }) => {
			const splitScope = OAuth2Clients.splitScope(scope || '');
			return normalized.every((item) => splitScope.includes(item));
		})?.length;
	}

	static async saveConsent(subject: User, client: OAuth2Client, scopes: string | string[]) {
		const normalized = OAuth2Clients.splitScope(scopes);
		const [existing] = await DB.drizzle
			.select()
			.from(oauth2ClientAuthorization)
			.where(
				and(
					eq(oauth2ClientAuthorization.clientId, client.id),
					eq(oauth2ClientAuthorization.userId, subject.id)
				)
			)
			.limit(1);

		if (existing) {
			const splitScope = OAuth2Clients.splitScope(existing.scope || '');
			normalized.forEach((entry) => {
				if (!splitScope.includes(entry)) {
					splitScope.push(entry);
				}
			});

			await DB.drizzle
				.update(oauth2ClientAuthorization)
				.set({ scope: OAuth2Clients.joinScope(splitScope), current: 1, expires_at: null })
				.where(eq(oauth2ClientAuthorization.id, existing.id));
			return;
		}

		await DB.drizzle.insert(oauth2ClientAuthorization).values({
			userId: subject.id,
			clientId: client.id,
			scope: OAuth2Clients.joinScope(normalized)
		});
	}

	static async revokeConsent(subject: User, clientId: string) {
		const client = await OAuth2Clients.fetchById(clientId);
		if (!client) return false;

		await OAuth2Tokens.wipeClientTokens(client, subject);
		await DB.drizzle
			.update(oauth2ClientAuthorization)
			.set({ current: 0, expires_at: new Date() })
			.where(
				and(
					eq(oauth2ClientAuthorization.userId, subject.id),
					eq(oauth2ClientAuthorization.clientId, client.id),
					eq(oauth2ClientAuthorization.current, 1)
				)
			);

		return true;
	}

	static async listAuthorizations(subject: User) {
		return DB.drizzle
			.select()
			.from(oauth2Client)
			.innerJoin(oauth2ClientAuthorization, eq(oauth2ClientAuthorization.clientId, oauth2Client.id))
			.leftJoin(oauth2ClientUrl, eq(oauth2ClientUrl.clientId, oauth2Client.id))
			.where(
				and(
					eq(oauth2ClientAuthorization.userId, subject.id),
					eq(oauth2ClientAuthorization.current, 1)
				)
			);
	}

	static async issueIdToken(subject: User, client: OAuth2Client, scope: string[], nonce?: string) {
		const userData: Record<string, unknown> = {
			name: subject.display_name,
			preferred_username: subject.username,
			nickname: subject.display_name,
			updated_at: subject.updated_at,
			nonce
		};

		if (scope.includes('email')) {
			userData.email = subject.email;
			userData.email_verified = true;
		}

		if (scope.includes('picture') && subject.pictureId) {
			userData.picture = `${env.PUBLIC_URL}/api/avatar/${subject.uuid}`;
		}

		return JWT.issue(userData, subject.uuid, client.client_id);
	}
}
