import { eq } from 'drizzle-orm';
import {
	DB,
	oauth2Client,
	upload,
	user,
	type OAuth2Client,
	type Upload,
	type User
} from './drizzle';
import { Users } from './users';
import { readFile, stat, unlink, writeFile } from 'fs/promises';
import { join } from 'path';
import * as mime from 'mime-types';
import { OAuth2Clients } from './oauth2';

export class Uploads {
	static userFallbackImage: Buffer;
	static clientFallbackImage: Buffer;
	static uploads = join('uploads');

	static async determineStaticPath() {
		try {
			await stat('static');
			return 'static';
		} catch {
			return join('build', 'client');
		}
	}

	static async getUserFallback() {
		if (!Uploads.userFallbackImage) {
			const staticPath = await Uploads.determineStaticPath();
			Uploads.userFallbackImage = await readFile(join(staticPath, 'avatar.png'));
		}
		return Uploads.userFallbackImage;
	}

	static async getClientFallback() {
		if (!Uploads.clientFallbackImage) {
			const staticPath = await Uploads.determineStaticPath();
			Uploads.clientFallbackImage = await readFile(join(staticPath, 'application.png'));
		}
		return Uploads.clientFallbackImage;
	}

	static async removeUpload(subject: Upload) {
		try {
			unlink(join(Uploads.uploads, subject.file));
		} catch {
			// ignore unlink error
		}

		await DB.drizzle.delete(upload).where(eq(upload.id, subject.id));
	}

	static async getAvatarByUuid(
		uuid: string
	): Promise<{ file: string; mimetype: string } | undefined> {
		const user = await Users.getByUuid(uuid);
		if (!user?.pictureId) {
			return undefined;
		}

		const [picture] = await DB.drizzle
			.select({ mimetype: upload.mimetype, file: upload.file })
			.from(upload)
			.where(eq(upload.id, user.pictureId));
		return picture;
	}

	static async getClientAvatarById(
		id: string
	): Promise<{ file: string; mimetype: string } | undefined> {
		const client = await OAuth2Clients.fetchById(id);
		if (!client?.pictureId) {
			return undefined;
		}

		const [picture] = await DB.drizzle
			.select({ mimetype: upload.mimetype, file: upload.file })
			.from(upload)
			.where(eq(upload.id, client.pictureId));
		return picture;
	}

	static async removeAvatar(subject: User) {
		if (!subject.pictureId) return;

		const [fileinfo] = await DB.drizzle
			.select()
			.from(upload)
			.where(eq(upload.id, subject.pictureId));
		if (fileinfo) {
			await Uploads.removeUpload(fileinfo);
		}

		await DB.drizzle.update(user).set({ pictureId: null }).where(eq(user.id, subject.id));
	}

	static async removeClientAvatar(client: OAuth2Client) {
		if (!client.pictureId) return;

		const [fileinfo] = await DB.drizzle
			.select()
			.from(upload)
			.where(eq(upload.id, client.pictureId));
		if (fileinfo) {
			await Uploads.removeUpload(fileinfo);
		}

		await DB.drizzle
			.update(oauth2Client)
			.set({ pictureId: null })
			.where(eq(oauth2Client.id, client.id));
	}

	static async saveAvatar(subject: User, file: File) {
		const ext = mime.extension(file.type);
		const newName = `user-${subject.uuid.split('-')[0]}-${Math.floor(Date.now() / 1000)}.${ext}`;
		const arrayBuffer = await file.arrayBuffer();
		// Write to filesystem
		await writeFile(join(Uploads.uploads, newName), Buffer.from(arrayBuffer));
		// Remove old
		await Uploads.removeAvatar(subject);
		// Update DB
		const [retval] = await DB.drizzle.insert(upload).values({
			original_name: file.name,
			mimetype: file.type,
			file: newName,
			uploaderId: subject.id
		});
		await DB.drizzle
			.update(user)
			.set({ pictureId: retval.insertId })
			.where(eq(user.id, subject.id));
	}

	static async saveClientAvatar(client: OAuth2Client, uploader: User, file: File) {
		const ext = mime.extension(file.type);
		const newName = `client-${client.client_id.substring(0, 8)}-${Math.floor(Date.now() / 1000)}.${ext}`;
		const arrayBuffer = await file.arrayBuffer();
		// Write to filesystem
		await writeFile(join(Uploads.uploads, newName), Buffer.from(arrayBuffer));
		// Remove old
		await Uploads.removeClientAvatar(client);
		// Update DB
		const [retval] = await DB.drizzle.insert(upload).values({
			original_name: file.name,
			mimetype: file.type,
			file: newName,
			uploaderId: uploader.id
		});
		await DB.drizzle
			.update(oauth2Client)
			.set({ pictureId: retval.insertId })
			.where(eq(oauth2Client.id, client.id));
	}
}
