import { asc, count, eq, like, or, sql } from 'drizzle-orm';
import {
	DB,
	privilege,
	user,
	userPrivilegesPrivilege,
	type Privilege,
	type User
} from '../drizzle';
import type { Paginated, PaginationMeta } from '$lib/types';

export interface AdminUserListItem extends Omit<User, 'password'> {
	privileges: Privilege[];
}

export class UsersAdmin {
	/**
	 * Map a list of users with their privileges.
	 * @param junkList User-privilege amalgamation
	 * @returns List of users
	 */
	static mergeUserResponse(
		junkList: {
			user: User;
			user_privileges_privilege?: typeof userPrivilegesPrivilege.$inferSelect | null;
			privilege?: Privilege | null;
		}[]
	) {
		return junkList.reduce<AdminUserListItem[]>((accum, dbe) => {
			let user = accum.find((entry) => entry.id === dbe.user.id);
			if (!user) {
				user = { ...dbe.user, password: undefined, privileges: [] } as AdminUserListItem;
				accum.push(user);
			}

			// Individual privilege
			if (dbe.user_privileges_privilege && dbe.privilege) {
				if (
					!user.privileges.some((priv) => priv.id === dbe.user_privileges_privilege?.privilegeId)
				) {
					user.privileges.push(dbe.privilege);
				}
			}

			return accum;
		}, []);
	}

	/**
	 * Get a list of all users and their privileges.
	 * @param params Search params
	 * @returns List of users
	 */
	static async getAllUsers({
		filter,
		offset = 0,
		limit = 20
	}: {
		filter?: string;
		offset?: number;
		limit?: number;
	}) {
		const subfilter = `%${filter}%`;
		const searchExpression = filter
			? or(
					like(user.uuid, subfilter),
					sql`lower(${user.username}) like ${subfilter}`,
					sql`lower(${user.display_name}) like ${subfilter}`,
					sql`lower(${user.email}) like ${subfilter}`
				)
			: undefined;

		const [{ rowCount }] = await DB.drizzle
			.select({ rowCount: count(user.id).mapWith(Number) })
			.from(user)
			.where(searchExpression);

		// LEFT JOINs below will include more rows than we allow with LIMIT,
		// so we need to do a subquery with the limiting first.
		const baseQuery = DB.drizzle
			.select({ id: user.id })
			.from(user)
			.where(searchExpression)
			.orderBy(asc(user.id))
			.limit(limit)
			.offset(offset)
			.as('searchBase');

		const junkList = await DB.drizzle
			.select({
				user: user,
				user_privileges_privilege: userPrivilegesPrivilege,
				privilege: privilege
			})
			.from(baseQuery)
			.innerJoin(user, eq(baseQuery.id, user.id))
			.leftJoin(userPrivilegesPrivilege, eq(userPrivilegesPrivilege.userId, user.id))
			.leftJoin(privilege, eq(userPrivilegesPrivilege.privilegeId, privilege.id))
			.where(searchExpression);

		const meta: PaginationMeta = {
			rowCount,
			pageSize: limit,
			pageCount: Math.ceil(rowCount / limit)
		};

		const list = UsersAdmin.mergeUserResponse(junkList);

		return <Paginated<AdminUserListItem>>{
			list,
			meta
		};
	}

	/**
	 * Get an user's information by their UUID.
	 * @param uuid User UUID
	 * @returns User infor
	 */
	static async getUserDetails(uuid: string) {
		const junkList = await DB.drizzle
			.select()
			.from(user)
			.leftJoin(userPrivilegesPrivilege, eq(userPrivilegesPrivilege.userId, user.id))
			.leftJoin(privilege, eq(userPrivilegesPrivilege.privilegeId, privilege.id))
			.where(eq(user.uuid, uuid));
		const [userInfo] = UsersAdmin.mergeUserResponse(junkList);
		return userInfo;
	}
}
