import bcrypt from 'bcryptjs';
import { and, eq, inArray, isNull, or, sql } from 'drizzle-orm';
import { DB, privilege, user, userPrivilegesPrivilege, type User } from '../drizzle';
import type { UserSession } from './types';
import { error, redirect } from '@sveltejs/kit';
import { CryptoUtils } from '../crypto-utils';
import { env as privateEnv } from '$env/dynamic/private';
import { Emails, ForgotPasswordEmail, InvitationEmail, RegistrationEmail } from '../email';
import { env as publicEnv } from '$env/dynamic/public';
import { UserTokens } from './tokens';

export class Users {
	/**
	 * Get user by their primary key ID.
	 * @param id User ID
	 * @returns User
	 */
	static async getById(id: number): Promise<User | undefined> {
		const [result] = await DB.drizzle
			.select()
			.from(user)
			.where(and(eq(user.id, id), eq(user.activated, 1)))
			.limit(1);
		return result;
	}

	/**
	 * Get user by their UUID (public-facing ID).
	 * @param id User ID
	 * @returns User
	 */
	static async getByUuid(uuid: string, activatedCheck = true): Promise<User | undefined> {
		const [result] = await DB.drizzle
			.select()
			.from(user)
			.where(and(eq(user.uuid, uuid), activatedCheck ? eq(user.activated, 1) : undefined))
			.limit(1);
		return result;
	}

	/**
	 * Get an user by their login (username or email)
	 * @param login Username or email address
	 * @returns User
	 */
	static async getByLogin(login: string): Promise<User | undefined> {
		const [result] = await DB.drizzle
			.select()
			.from(user)
			.where(
				and(
					or(eq(user.email, login), sql`lower(${user.username}) = ${login.toLowerCase()}`),
					eq(user.activated, 1)
				)
			)
			.limit(1);
		return result;
	}

	/**
	 * Get an user by their session cookie.
	 * @param session Session from cookie
	 * @returns User
	 */
	static async getBySession(session?: UserSession): Promise<User | undefined> {
		if (!session) return undefined;
		const [result] = await DB.drizzle
			.select()
			.from(user)
			.where(and(eq(user.id, session.uid), eq(user.activated, 1)))
			.limit(1);
		return result;
	}

	/**
	 * Update an user.
	 * @param subject User to update
	 * @param fields Fields to set
	 */
	static async update(subject: User, fields: Partial<User>) {
		return DB.drizzle.update(user).set(fields).where(eq(user.id, subject.id));
	}

	/**
	 * Check an user's password.
	 * @param user User
	 * @param password Plaintext password
	 * @returns Boolean
	 */
	static async validatePassword(user: User, password: string): Promise<boolean> {
		return bcrypt.compare(password, user.password as string);
	}

	/**
	 * Hash a password.
	 * @param password Plaintext password
	 * @returns Hashed password
	 */
	static async hashPassword(password: string): Promise<string> {
		return bcrypt.hash(password, 10);
	}

	/**
	 * Create a session cookie from user information.
	 * @param user User
	 * @returns Cookie
	 */
	static async toSession(user: User): Promise<UserSession> {
		return {
			uid: user.id,
			uuid: user.uuid,
			name: user.display_name,
			username: user.username
		};
	}

	/**
	 * Get an user from the request locals or redirect to the login page.
	 * @param locals App locals
	 * @param url URL
	 * @returns User
	 */
	static async readSessionOrRedirect(locals: App.Locals, url: URL) {
		const currentUser = await Users.getBySession(locals.session.data?.user);
		if (!currentUser) {
			await locals.session.destroy();
			return redirect(301, `/login?redirectTo=${encodeURIComponent(url.pathname + url.search)}`);
		}
		return currentUser;
	}

	/**
	 * Check if an user exists with the provided username and/or email address.
	 * @param username Username
	 * @param email Email address
	 * @returns true if existing user does not exist
	 */
	static async checkRegistration(username: string, email: string) {
		return !(
			await DB.drizzle
				.select({ id: user.id })
				.from(user)
				.where(
					or(
						sql`lower(${user.username}) = ${username.toLowerCase()}`,
						sql`lower(${user.email}) = ${email.toLowerCase()}`
					)
				)
		)?.length;
	}

	/**
	 * Get user information by activation token.
	 * @param token Activation token
	 * @returns User information
	 */
	static async getActivationToken(token: string) {
		const returnedToken = await UserTokens.getByToken(token, 'activation');
		if (!returnedToken?.userId) return undefined;

		const [userInfo] = await DB.drizzle
			.select()
			.from(user)
			.where(eq(user.id, returnedToken.userId as number));

		return {
			...returnedToken,
			user: userInfo
		};
	}

	/**
	 * Activate an user by an activation token.
	 * @param token Token
	 * @param subject User
	 */
	static async activateUserBy(token: string, subject: User) {
		await DB.drizzle
			.update(user)
			.set({ activated: 1, activity_at: new Date() })
			.where(eq(user.id, subject.id));
		await UserTokens.remove(token);
	}

	/**
	 * Register a new user.
	 * @param details User details
	 * @returns New User
	 */
	static async register({
		username,
		displayName,
		email,
		password,
		activate = false
	}: {
		username: string;
		displayName: string;
		email: string;
		password: string;
		activate?: boolean;
	}) {
		const passwordHash = await Users.hashPassword(password);
		const [retval] = await DB.drizzle.insert(user).values({
			uuid: CryptoUtils.createUUID(),
			email,
			username,
			password: passwordHash,
			display_name: displayName,
			activated: privateEnv.EMAIL_ENABLED === 'false' ? 1 : Number(activate),
			activity_at: new Date()
		});

		const [newUser] = await DB.drizzle.select().from(user).where(eq(user.id, retval.insertId));

		if (privateEnv.EMAIL_ENABLED !== 'false' && !activate) {
			await Users.sendRegistrationEmail(newUser);
		}

		// TODO: audit log

		return newUser;
	}

	/**
	 * Send an activation email to an user.
	 * @param user User
	 */
	static async sendRegistrationEmail(user: User) {
		const token = await UserTokens.create(
			'activation',
			new Date(Date.now() + 7 * 24 * 60 * 60 * 1000),
			user.id
		);

		const params = new URLSearchParams({ activate: token.token });
		const content = RegistrationEmail(
			user.username,
			`${publicEnv.PUBLIC_URL}/login?${params.toString()}`
		);

		// TODO: logging
		try {
			await Emails.getSender().sendTemplate(
				user.email,
				`Activate your account on ${publicEnv.PUBLIC_SITE_NAME}`,
				content
			);
		} catch (error) {
			console.error(error);
			await UserTokens.remove(token);
		}
	}

	/**
	 * Send a password reset email to an user.
	 * @param user User
	 */
	static async sendPasswordEmail(user: User) {
		const token = await UserTokens.create(
			'password',
			new Date(Date.now() + 60 * 60 * 1000),
			user.id
		);
		const params = new URLSearchParams({ token: token.token });
		const content = ForgotPasswordEmail(
			user.username,
			`${publicEnv.PUBLIC_URL}/login/password?${params.toString()}`
		);

		// TODO: logging
		try {
			await Emails.getSender().sendTemplate(
				user.email,
				`Reset your password on ${publicEnv.PUBLIC_SITE_NAME}`,
				content
			);
		} catch {
			await UserTokens.remove(token);
		}
	}

	/**
	 * Send a registration invite to an email address.
	 * @param email Email to invite
	 */
	static async sendInvitationEmail(email: string) {
		const token = await UserTokens.create(
			'invite',
			new Date(Date.now() + 7 * 24 * 60 * 60 * 1000),
			undefined,
			undefined,
			`register=${email}`
		);
		const params = new URLSearchParams({ token: token.token });
		const content = InvitationEmail(`${publicEnv.PUBLIC_URL}/register?${params.toString()}`);

		// TODO: logging
		try {
			await Emails.getSender().sendTemplate(
				email,
				`You have been invited to create an account on ${publicEnv.PUBLIC_SITE_NAME}`,
				content
			);
		} catch {
			await UserTokens.remove(token);
		}
	}

	/**
	 * Get all available privileges.
	 * @param clientId (optional)	Client privileges
	 * @returns Available privileges
	 */
	static async getAvailablePrivileges(clientId?: number) {
		return await DB.drizzle
			.select()
			.from(privilege)
			.where(clientId ? eq(privilege.clientId, clientId) : isNull(privilege.clientId));
	}

	/**
	 * Get an user's privileges.
	 * @param subject User
	 * @param clientId (optional) OAuth2 client ID, returns privileges for a specific client only
	 * @returns User privileges (string list)
	 */
	static async getUserPrivileges(subject: User, clientId?: number) {
		const list = await DB.drizzle
			.select({
				privilege: privilege.name
			})
			.from(privilege)
			.innerJoin(userPrivilegesPrivilege, eq(privilege.id, userPrivilegesPrivilege.privilegeId))
			.where(
				and(
					eq(userPrivilegesPrivilege.userId, subject.id),
					clientId ? eq(privilege.clientId, clientId) : isNull(privilege.clientId)
				)
			);

		return list.reduce<string[]>(
			(accum, { privilege }) => (!accum.includes(privilege) ? [...accum, privilege] : accum),
			[]
		);
	}

	/**
	 * Grant a new privilege to a user.
	 *
	 * **This can only be used to grant system privileges.**
	 * @see setUserPrivileges for general privilege changes.
	 * @param subject Recipient
	 * @param name Privilege name
	 * @returns Boolean, whether the privilege was granted or not.
	 */
	static async grantPrivilege(subject: User, name: string) {
		const [existingPrivilege] = await DB.drizzle
			.select({ id: privilege.id })
			.from(privilege)
			.where(and(eq(privilege.name, name), isNull(privilege.clientId)));
		if (!existingPrivilege) return false;

		const [alreadyHas] = await DB.drizzle
			.select({ privilegeId: userPrivilegesPrivilege.privilegeId })
			.from(userPrivilegesPrivilege)
			.where(
				and(
					eq(userPrivilegesPrivilege.userId, subject.id),
					eq(userPrivilegesPrivilege.privilegeId, existingPrivilege.id)
				)
			);
		if (alreadyHas) return true;

		await DB.drizzle.insert(userPrivilegesPrivilege).values({
			privilegeId: existingPrivilege.id,
			userId: subject.id
		});

		return true;
	}

	/**
	 * Set an user's privileges to an array of IDs.
	 * @param subject User
	 * @param privilegeIds Privileges
	 * @param clientId (optional) OAuth2 client ID, updates privileges related to a specific client only.
	 */
	static async setUserPrivileges(subject: User, privilegeIds: number[], clientId?: number) {
		// Prevent tricksters from trying to give themselves system privileges
		// through the OAuth2 client privilege system. (or remove them from others)
		// The privileges in question must actually be related to the specified client.
		if (clientId) {
			for (const id of privilegeIds) {
				const [exists] = await DB.drizzle
					.select({ id: privilege.id })
					.from(privilege)
					.where(and(eq(privilege.id, id), eq(privilege.clientId, clientId)));
				if (!exists) {
					// TODO: logging
					return error(403);
				}
			}
		}

		const current = await DB.drizzle
			.select({
				privilegeId: userPrivilegesPrivilege.privilegeId
			})
			.from(userPrivilegesPrivilege)
			.innerJoin(privilege, eq(privilege.id, userPrivilegesPrivilege.privilegeId))
			.where(
				and(
					eq(userPrivilegesPrivilege.userId, subject.id),
					clientId ? eq(privilege.clientId, clientId) : undefined
				)
			);

		const toRemoveIds = current.reduce<number[]>(
			(list, { privilegeId }) =>
				!privilegeIds.includes(privilegeId) ? [...list, privilegeId] : list,
			[]
		);

		if (toRemoveIds.length) {
			await DB.drizzle
				.delete(userPrivilegesPrivilege)
				.where(
					and(
						eq(userPrivilegesPrivilege.userId, subject.id),
						inArray(userPrivilegesPrivilege.privilegeId, toRemoveIds)
					)
				);
		}

		const toInsertIds = privilegeIds.reduce<number[]>(
			(list, id) => (!current.some(({ privilegeId }) => privilegeId === id) ? [...list, id] : list),
			[]
		);

		if (toInsertIds.length) {
			await DB.drizzle
				.insert(userPrivilegesPrivilege)
				.values(toInsertIds.map((privilegeId) => ({ userId: subject.id, privilegeId })));
		}
	}

	/**
	 * Hide an email addresses username part partially as a hint.
	 * @param email Email address
	 * @returns Slightly redacted email address
	 */
	static anonymizeEmail(email: string) {
		const [name, domain] = email.split('@');
		const namePart = `${name.charAt(0)}${''.padStart(name.length - 2, '*')}${name.charAt(name.length - 1)}`;
		return `${namePart}@${domain}`;
	}
}

export * from './totp';
export * from './types';
export * from './tokens';
