import { and, eq, gt, isNull, or, sql } from 'drizzle-orm';
import { CryptoUtils } from '../crypto-utils';
import { DB, userToken, type User, type UserToken } from '../drizzle';

export class UserTokens {
	static async create(
		type: (typeof userToken.$inferInsert)['type'],
		expires: Date,
		userId?: number,
		nonce?: string,
		metadata?: string
	) {
		const token = CryptoUtils.generateString(64);
		const obj = <typeof userToken.$inferInsert>{
			type,
			token,
			userId,
			expires_at: expires,
			nonce,
			metadata
		};
		const [retval] = await DB.drizzle.insert(userToken).values(obj);
		return { id: retval.insertId, ...obj } as UserToken;
	}

	static async remove(token: string | { token: string }) {
		const removeBy = typeof token === 'string' ? token : token.token;
		await DB.drizzle.delete(userToken).where(eq(userToken.token, removeBy));
	}

	static async getByToken(token: string, type: (typeof userToken.$inferSelect)['type']) {
		const [returned] = await DB.drizzle
			.select()
			.from(userToken)
			.where(
				and(
					eq(userToken.token, token),
					eq(userToken.type, type),
					or(isNull(userToken.expires_at), gt(userToken.expires_at, new Date()))
				)
			)
			.limit(1);
		return returned;
	}

	static async wipeUserTokens(user: User) {
		await DB.drizzle.delete(userToken).where(eq(userToken.userId, user.id));
	}

	static async wipeExpiredTokens() {
		await DB.drizzle.execute(sql`DELETE FROM ${userToken} WHERE ${userToken.expires_at} < NOW()`);
	}
}
