import { Challenge } from '$lib/server/challenge.js';
import type { User } from '$lib/server/drizzle';
import { Uploads } from '$lib/server/upload.js';
import { Users, type UserSession } from '$lib/server/users/index.js';
import { TimeOTP } from '$lib/server/users/totp.js';
import { passwordRegex } from '$lib/validators.js';
import { fail, redirect } from '@sveltejs/kit';

interface AccountUpdate {
	displayName: string;

	currentEmail: string;
	newEmail: string;

	currentPassword: string;
	newPassword: string;
}

export const actions = {
	logout: async ({ locals }) => {
		await locals.session.destroy();
		return redirect(303, '/');
	},
	update: async ({ request, locals }) => {
		const currentUser = await Users.getBySession(locals.session.data?.user);
		if (!currentUser) {
			await locals.session.destroy();
			return redirect(303, '/login');
		}

		const body = await request.formData();
		let data: Partial<AccountUpdate>;
		try {
			data = await Challenge.authorizedChanges<AccountUpdate>(
				['displayName', 'currentEmail', 'newEmail', 'currentPassword', 'newPassword'],
				body,
				currentUser
			);
		} catch {
			return fail(400, { errors: ['invalidRequest'] });
		}

		// No changes
		if (
			(!data.displayName || data.displayName === currentUser.display_name) &&
			!data.newEmail &&
			!data.newPassword
		) {
			return { success: true, errors: <string[]>[], fields: <string[]>[] };
		}

		// Current email is not provided
		if (data.newEmail && !data.currentEmail) {
			return fail(400, {
				displayName: data.displayName,
				errors: ['emailRequired'],
				fields: ['currentEmail']
			});
		}

		// Current email is invalid
		if (data.currentEmail && data.currentEmail !== currentUser.email) {
			return fail(400, {
				displayName: data.displayName,
				errors: ['invalidEmail'],
				fields: ['currentEmail']
			});
		}

		// Current password is not provided
		if (data.newPassword && !data.currentPassword) {
			return fail(400, {
				displayName: data.displayName,
				errors: ['passwordRequired'],
				fields: ['currentPassword']
			});
		}

		// Password does not match current password
		if (
			data.currentPassword &&
			!(await Users.validatePassword(currentUser, data.currentPassword))
		) {
			return fail(400, {
				displayName: data.displayName,
				errors: ['invalidPassword'],
				fields: ['currentPassword']
			});
		}

		if (data.newPassword && !passwordRegex.test(data.newPassword)) {
			return fail(400, {
				displayName: data.displayName,
				errors: ['invalidPassword'],
				fields: ['newPassword']
			});
		}

		// Invalid display name
		if (data.displayName && (data.displayName.length < 3 || data.displayName.length > 32)) {
			return fail(400, {
				displayName: data.displayName,
				errors: ['invalidDisplayName'],
				fields: ['displayName']
			});
		}

		// When updating email or password, we check if OTP has been enabled.
		// If it is, we need to ask for the OTP code.
		if (!body.has('challenge') && (data.newEmail || data.newPassword)) {
			const isOtp = await TimeOTP.isUserOtp(currentUser);
			if (isOtp) {
				const challenge = await Challenge.issueChallenge(data, currentUser.uuid);
				return { otpRequired: challenge };
			}
		}

		// Update the user table
		const updates: Partial<User> = {};
		if (data.displayName) {
			updates.display_name = data.displayName;
		}

		if (data.newEmail) {
			updates.email = data.newEmail;
		}

		if (data.newPassword) {
			updates.password = await Users.hashPassword(data.newPassword);
		}

		await Users.update(currentUser, updates);

		// Update session display name
		if (data.displayName) {
			await locals.session.update(({ user }) => ({
				user: { ...user, name: data.displayName as string } as UserSession
			}));
		}

		// TODO: audit log

		return {
			success: true,
			errors: <string[]>[],
			fields: <string[]>[],
			displayName: data.displayName || currentUser.display_name
		};
	},
	avatar: async ({ request, locals }) => {
		const currentUser = await Users.getBySession(locals.session.data?.user);
		if (!currentUser) {
			await locals.session.destroy();
			return redirect(303, '/login');
		}

		const formData = Object.fromEntries(await request.formData());
		if (!(formData.file as File)?.name || (formData.file as File).name === 'undefined') {
			return fail(400, {
				error: true,
				message: 'You must provide a file to upload'
			});
		}

		const { file } = formData as { file: File };

		await Uploads.saveAvatar(currentUser, file);

		return { avatarChanged: true };
	},
	removeAvatar: async ({ locals }) => {
		const currentUser = await Users.getBySession(locals.session.data?.user);
		if (!currentUser) {
			await locals.session.destroy();
			return redirect(303, '/login');
		}

		await Uploads.removeAvatar(currentUser);

		return { avatarChanged: true };
	}
};

export async function load({ locals, url }) {
	const userInfo = locals.session.data?.user;
	const currentUser = await Users.getBySession(userInfo);
	if (!userInfo || !currentUser) {
		await locals.session.destroy();
		return redirect(301, `/login?redirectTo=${encodeURIComponent(url.pathname)}`);
	}

	const privileges = await Users.getUserPrivileges(currentUser);
	const otpEnabled = await TimeOTP.isUserOtp(currentUser);
	const updateRef = Date.now();

	return {
		privileges,
		user: userInfo,
		email: Users.anonymizeEmail(currentUser.email),
		otpEnabled,
		hasAvatar: !!currentUser.pictureId,
		updateRef
	};
}
