import { Challenge, type ChallengeBody } from '$lib/server/challenge.js';
import { Changesets } from '$lib/server/changesets.js';
import { CryptoUtils } from '$lib/server/crypto-utils';
import type { User } from '$lib/server/drizzle';
import { UserTokens, Users } from '$lib/server/users';
import { TimeOTP } from '$lib/server/users/totp.js';
import { fail, redirect } from '@sveltejs/kit';
import * as QRCode from 'qrcode';

interface ActivateChallenge {
	secret: string;
}

interface ActivateRequest {
	action: 'deactivate' | 'activate';
	challenge: string;
	otpCode?: string;
}

const issueActivateChallenge = async (subject: User) => {
	const secret = TimeOTP.createSecret();
	const challenge = await Challenge.issueChallenge<ActivateChallenge>({ secret }, subject.uuid);
	const uri = TimeOTP.getUri(secret, subject.username);
	const qr = await QRCode.toDataURL(uri);
	return { challenge, qr, uri, action: 'activate' };
};

export const actions = {
	activate: async ({ locals, request }) => {
		const currentUser = await Users.getBySession(locals.session.data?.user);
		if (!currentUser) {
			await locals.session.destroy();
			return redirect(303, '/login');
		}

		const body = await request.formData();
		const { challenge, otpCode } = Changesets.take<ActivateRequest>(['challenge', 'otpCode'], body);

		if (!challenge) {
			return issueActivateChallenge(currentUser);
		}

		const decoded = await CryptoUtils.decryptChallenge<ChallengeBody<ActivateChallenge>>(challenge);

		if (
			!otpCode ||
			decoded?.exp < Date.now() ||
			decoded?.aud !== currentUser.uuid ||
			!decoded?.data?.secret ||
			!TimeOTP.validate(decoded.data.secret, otpCode)
		) {
			return fail(400, { invalid: true, action: 'activate' });
		}

		await TimeOTP.saveUserOtp(currentUser, decoded.data.secret);

		// TODO: audit log

		return { success: true, action: 'activate' };
	},
	deactivate: async ({ request, locals }) => {
		const currentUser = await Users.getBySession(locals.session.data?.user);
		if (!currentUser) {
			await locals.session.destroy();
			return redirect(303, '/login');
		}

		const body = await request.formData();
		const { challenge, otpCode } = Changesets.take<ActivateRequest>(['challenge', 'otpCode'], body);

		const userOtp = await TimeOTP.getUserOtp(currentUser);
		if (!userOtp) {
			// Invalid request, no OTP activated...
			return redirect(303, '/account');
		}

		const secretHash = CryptoUtils.sha256hash(userOtp.token).toString('hex');

		if (!challenge) {
			const issued = await Challenge.issueChallenge<ActivateChallenge>(
				{ secret: secretHash },
				currentUser.uuid
			);
			return { challenge: issued, action: 'deactivate' };
		}

		const decoded = await CryptoUtils.decryptChallenge<ChallengeBody<ActivateChallenge>>(challenge);

		if (
			!otpCode ||
			decoded?.exp < Date.now() ||
			decoded?.aud !== currentUser.uuid ||
			decoded?.data?.secret !== secretHash ||
			!TimeOTP.validate(userOtp.token, otpCode)
		) {
			return fail(400, { invalid: true, action: 'deactivate' });
		}

		await UserTokens.remove(userOtp);

		// TODO: audit log

		return { success: true, action: 'deactivate' };
	}
};

export async function load({ locals }) {
	const currentUser = await Users.getBySession(locals.session.data?.user);
	if (!currentUser) {
		await locals.session.destroy();
		return redirect(301, '/login');
	}

	const otpEnabled = await TimeOTP.isUserOtp(currentUser);
	return {
		otpEnabled
	};
}
