import { env } from '$env/dynamic/public';
import type { User } from '$lib/server/drizzle/schema.js';
import { OAuth2BearerController } from '$lib/server/oauth2/controller/bearer.js';
import { AccessDenied, OAuth2Error } from '$lib/server/oauth2/error.js';
import { OAuth2Clients } from '$lib/server/oauth2/model/client.js';
import { OAuth2Response } from '$lib/server/oauth2/response.js';
import { Users } from '$lib/server/users/index.js';
import { json } from '@sveltejs/kit';

export const GET = async ({ request, url, locals }) => {
	let user: User | undefined = undefined;
	let clientId: number | undefined = undefined;
	let tokenScopes: string[] | undefined = undefined;

	if (locals.session?.data?.user) {
		// From session
		user = await Users.getBySession(locals.session.data.user);
	} else {
		// From OAuth2 bearer token
		try {
			const token = await OAuth2BearerController.bearerFromRequest(request, url);
			if (token?.userId) {
				tokenScopes = OAuth2Clients.splitScope(token.scope || '');
				user = await Users.getById(token.userId);
				clientId = token.clientId || undefined;
			}
		} catch (error) {
			if (error instanceof OAuth2Error) {
				return OAuth2Response.createErrorResponse(error);
			}
			throw error;
		}
	}

	if (!user) {
		return OAuth2Response.createErrorResponse(new AccessDenied('Access denied'));
	}

	const scopelessAccess = !tokenScopes || tokenScopes.includes('management');

	const userData: Record<string, unknown> = {
		id: user.id,
		uuid: user.uuid,
		username: user.username,
		display_name: user.display_name,

		// Standard claims
		sub: user.uuid,
		name: user.display_name,
		preferred_username: user.username,
		nickname: user.display_name
	};

	if (scopelessAccess || tokenScopes?.includes('email')) {
		userData.email = user.email;
		userData.email_verified = true;
	}

	if ((scopelessAccess || tokenScopes?.includes('picture')) && user.pictureId) {
		userData.picture = `${env.PUBLIC_URL}/api/avatar/${user.uuid}`;
	}

	if (scopelessAccess || tokenScopes?.includes('privileges')) {
		userData.privileges = await Users.getUserPrivileges(user, clientId);
	}

	return json(userData);
};
