import { OAuth2Clients, OAuth2DeviceCodes, OAuth2Users } from '$lib/server/oauth2/index.js';
import { Users } from '$lib/server/users';
import { error, fail, redirect } from '@sveltejs/kit';
import { RateLimiter } from 'sveltekit-rate-limiter/server';

const limiter = new RateLimiter({
	IP: [6, 'm']
});

export const actions = {
	default: async (event) => {
		const { locals, request, url } = event;
		const currentUser = await Users.getBySession(locals.session.data?.user);
		if (!currentUser) {
			await locals.session.destroy();
			return redirect(301, `/login?redirectTo=${encodeURIComponent(url.pathname)}`);
		}

		const body = await request.formData();
		const code = body.get('code') as string;
		if (!body.has('code') || !code) {
			if (await limiter.isLimited(event)) throw error(429, "You're doing that too much!");
			return fail(400, { errors: ['noCode'] });
		}

		const token = await OAuth2DeviceCodes.getByUserCode(code);
		if (!token?.clientId) {
			if (await limiter.isLimited(event)) throw error(429, "You're doing that too much!");
			return fail(404, { errors: ['invalidCode'] });
		}

		const scopes = OAuth2Clients.transformScope(token.scope || '');
		const client = await OAuth2Clients.fetchById(token.clientId);

		// Due to the risk of this API being used with unauthenticated clients,
		// implicit consent is not allowed like it is with normal authorization flow.

		if (!body.has('decision')) {
			const clientInfo = await OAuth2Clients.authorizeClientInfo(client, scopes);

			return { code, client: clientInfo, errors: [] };
		}

		const consented = body.get('decision') === '1';

		const success = await OAuth2DeviceCodes.authorizeByUserCode(
			currentUser,
			body.get('code') as string,
			consented
		);

		if (!success) {
			return fail(404, { errors: ['invalidCode'] });
		}

		if (consented) {
			await OAuth2Users.saveConsent(currentUser, client, scopes);
		}

		return { errors: [] };
	}
};

export const load = async ({ locals, url }) => {
	const userInfo = locals.session.data?.user;
	const currentUser = await Users.getBySession(userInfo);
	if (!userInfo || !currentUser) {
		await locals.session.destroy();
		return redirect(301, `/login?redirectTo=${encodeURIComponent(url.pathname)}`);
	}

	return {
		user: userInfo
	};
};
