import { Changesets } from '$lib/server/changesets.js';
import { Users } from '$lib/server/users/index.js';
import { UserTokens } from '$lib/server/users/tokens.js';
import { emailRegex, passwordRegex } from '$lib/validators.js';
import { error, redirect } from '@sveltejs/kit';
import { RateLimiter } from 'sveltekit-rate-limiter/server';

interface PasswordRequest {
	newPassword: string;
	repeatPassword: string;
}

// Sending an email asynchronously has a similar amount of delay,
// so lets fake it. TODO: offload email sending somewhere else.
const failDelay = () =>
	new Promise((resolve) => setTimeout(resolve, 2000 + (Math.random() * 2000 - 1000)));

const limiter = new RateLimiter({
	IP: [3, '45s']
});

export const actions = {
	sendEmail: async (event) => {
		const { locals, request } = event;
		if (await limiter.isLimited(event)) throw error(429);
		if (locals.session.data?.user) {
			return redirect(303, '/');
		}

		const body = await request.formData();
		const { email } = Changesets.take<{ email: string }>(['email'], body);

		if (!email || !emailRegex.test(email)) {
			return { errors: ['invalidEmail'] };
		}

		const user = await Users.getByLogin(email);
		if (!user || user.activated === 0) {
			await failDelay();
			return { success: 'sent' };
		}

		try {
			await Users.sendPasswordEmail(user);
		} catch {
			// Ignore errors
			await failDelay();
		}

		return { success: 'sent' };
	},
	setPassword: async ({ locals, request, url }) => {
		if (locals.session.data?.user) {
			return redirect(303, '/');
		}

		const token = url.searchParams.get('token');
		if (!token) {
			return {
				errors: ['invalidPasswordToken']
			};
		}

		const exists = await UserTokens.getByToken(token, 'password');
		if (!exists?.userId) {
			return {
				errors: ['invalidPasswordToken']
			};
		}

		const user = await Users.getById(exists.userId as number);
		if (!user?.id || user.activated === 0) {
			return {
				errors: ['invalidPasswordToken']
			};
		}

		const body = await request.formData();
		const { newPassword, repeatPassword } = Changesets.take<PasswordRequest>(
			['newPassword', 'repeatPassword'],
			body
		);

		if (!newPassword || !passwordRegex.test(newPassword)) {
			return { errors: ['invalidPassword'] };
		}

		if (newPassword !== repeatPassword) {
			return { errors: ['passwordMismatch'] };
		}

		if (await Users.validatePassword(user, newPassword)) {
			return { errors: ['passwordSame'] };
		}

		const hashed = await Users.hashPassword(newPassword);
		await Users.update(user, { password: hashed });
		await UserTokens.remove(exists);

		return { success: 'set' };
	}
};

export const load = async ({ locals, url }) => {
	if (locals.session.data?.user) {
		return redirect(301, '/');
	}

	const token = url.searchParams.get('token');
	if (token) {
		const exists = await UserTokens.getByToken(token, 'password');
		if (!exists) {
			return redirect(301, '/login/password');
		}

		return { setter: true };
	}

	return { setter: false };
};
