import { AdminUtils } from '$lib/server/admin-utils';
import { Changesets } from '$lib/server/changesets.js';
import { OAuth2Tokens } from '$lib/server/oauth2/index.js';
import { Uploads } from '$lib/server/upload.js';
import { UsersAdmin } from '$lib/server/users/admin.js';
import { UserTokens, Users } from '$lib/server/users/index.js';
import { TimeOTP } from '$lib/server/users/totp.js';
import { hasPrivileges } from '$lib/utils.js';
import { emailRegex } from '$lib/validators.js';
import { error, fail } from '@sveltejs/kit';

interface UpdateRequest {
	displayName?: string;
	email?: string;
	activated?: boolean;
	privileges?: string;
}

export const actions = {
	removeOtp: async () => {},
	removeAvatar: async ({ locals, params: { uuid } }) => {
		await AdminUtils.getActionUser(locals, ['admin', 'admin:user']);

		const targetUser = await Users.getByUuid(uuid, false);
		if (!targetUser) {
			return fail(404, { errors: ['invalidUuid'] });
		}

		await Uploads.removeAvatar(targetUser);

		return { errors: [] };
	},
	deleteInfo: async ({ locals, params: { uuid } }) => {
		await AdminUtils.getActionUser(locals, ['admin', 'admin:user']);

		const targetUser = await Users.getByUuid(uuid, false);
		if (!targetUser || !!targetUser.activated) {
			return fail(404, { errors: ['invalidUuid'] });
		}

		// TODO: audit log

		const [stubName] = uuid.split('-');

		// Nuke EVERYTHING
		await UserTokens.wipeUserTokens(targetUser);
		await OAuth2Tokens.wipeUserTokens(targetUser);
		await Uploads.removeAvatar(targetUser);
		await Users.update(targetUser, {
			username: `stub${stubName}`,
			display_name: `Stub ${stubName}`,
			email: `${stubName}@uuid-stub.target`,
			activated: 0,
			updated_at: new Date()
		});

		return { errors: [] };
	},
	email: async ({ locals, params: { uuid }, url }) => {
		await AdminUtils.getActionUser(locals, ['admin', 'admin:user']);

		const type = url.searchParams.get('type') as 'password' | 'activate';
		if (!type) {
			return fail(403, { errors: ['invalidEmailType'] });
		}

		const targetUser = await Users.getByUuid(uuid, false);
		if (!targetUser) {
			return fail(404, { errors: ['invalidUuid'] });
		}

		switch (type) {
			case 'password':
				await Users.sendPasswordEmail(targetUser);
				break;
			case 'activate':
				await Users.sendRegistrationEmail(targetUser);
				break;
			default:
				return fail(403, { errors: ['invalidEmailType'] });
		}

		// TODO: audit log

		return { errors: [] };
	},
	update: async ({ locals, params: { uuid }, request }) => {
		const { currentUser, userSession } = await AdminUtils.getActionUser(locals, [
			'admin',
			'admin:user'
		]);
		const body = await request.formData();

		const { displayName, email, activated, privileges } = Changesets.take<UpdateRequest>(
			['displayName', 'email', 'activated', 'privileges'],
			body
		);

		if (
			privileges !== undefined &&
			!hasPrivileges(userSession.privileges || [], ['admin:user:privilege'])
		) {
			return fail(403, { errors: ['unauthorized'] });
		}

		const targetUser = await Users.getByUuid(uuid, false);
		if (!targetUser) {
			return fail(404, { errors: ['invalidUuid'] });
		}

		if (currentUser.id === targetUser.id && !activated) {
			return fail(400, { errors: ['lockout'] });
		}

		if (privileges !== undefined) {
			const newPrivilegeIds =
				privileges?.split(',').reduce<number[]>((final, entry) => {
					if (!entry) return final;
					const parsed = Number(entry);
					if (isNaN(parsed) || final.includes(parsed)) return final;
					return [...final, parsed];
				}, []) || [];

			await Users.setUserPrivileges(targetUser, newPrivilegeIds);
		}

		if (displayName && (displayName.length < 3 || displayName.length > 32)) {
			return fail(400, { errors: ['invalidDisplayName'] });
		}

		if (email && !emailRegex.test(email)) {
			return fail(400, { errors: ['invalidEmail'] });
		}

		await Users.update(targetUser, {
			display_name: displayName,
			email,
			activated: Number(!!activated),
			updated_at: new Date()
		});

		// TODO: audit log

		return { errors: [] };
	}
};

export const load = async ({ parent, params: { uuid } }) => {
	const { user } = await parent();
	AdminUtils.checkPrivileges(user, ['admin:user']);

	const userInfo = await UsersAdmin.getUserDetails(uuid);
	if (!userInfo) {
		error(404, 'User not found');
	}

	const privilegeRight = hasPrivileges(user.privileges, ['admin:user:privilege']);
	const otpEnabled = await TimeOTP.isUserOtp(userInfo);
	const privileges = privilegeRight ? await Users.getAvailablePrivileges() : [];

	return {
		privilegeRight,
		privileges,
		details: {
			...userInfo,
			otpEnabled
		}
	};
};
